#include <stdio.h>
#include <stdint.h>
#include <errno.h>

/* romchu 0.1 */
/* an attempt at parsing the new-style (version 2?) romc */

int main(int argc, char **argv)
{
    FILE *infile = fopen(argv[1], "rb");
    unsigned char head_buf[4];
    unsigned char payload_buf[0x10000];
    int block_count = 0;
    uint64_t nominal_size;
    const uint32_t header_mult = 0x400000;
    const uint32_t block_mult = 0x10000;

    /* read header */
    fread(head_buf, 4, 1, infile);

    if (head_buf[3] != 2)
    {
        fprintf(stderr,"Expected version 2 romc\n");
        return 1;
    }

    nominal_size = head_buf[2];
    nominal_size *= 0x100;
    nominal_size |= head_buf[1];
    nominal_size *= 0x100;
    nominal_size |= head_buf[0];
    nominal_size *= header_mult;

    while (1 == fread(head_buf, 4, 1, infile))
    {
        uint32_t payload_bytes;
        int payload_bits;
        uint32_t read_size;

        printf("%08lx=%08lx: %02x\n",
            (unsigned long)(ftell(infile)-4),
            (unsigned long)block_count*block_mult,
            (unsigned int)head_buf[0]&0xF);

        if (head_buf[0] & 1)
        {
            uint32_t block_size, block_bits;
            /* compressed */

            /* bytes */
            block_size = head_buf[3];
            block_size *= 0x100;
            block_size |= head_buf[2];
            block_size *= 0x100;
            block_size |= head_buf[1];
            block_size *= 0x10;
            block_size |= head_buf[0]>>4;

            /* and bits */
            block_bits = (head_buf[0] >> 1) & 7;

            payload_bytes = block_size - 4;
            payload_bits = block_bits;
        }
        else
        {
            uint32_t block_size;
            /* uncompressed */

            /* bytes */
            block_size = head_buf[3];
            block_size *= 0x100;
            block_size |= head_buf[2];
            block_size *= 0x100;
            block_size |= head_buf[1];
            block_size *= 0x80;
            block_size |= head_buf[0]>>1;

            payload_bytes = block_size;
            payload_bits = 0;
        }

        /* read payload */
        read_size = payload_bytes;
        if (payload_bits > 0)
        {
            read_size ++;
        }
        if (read_size > sizeof(payload_buf))
        {
            fprintf(stderr, "payload too large\n");
            return 1;
        }
        if (1 != fread(payload_buf, read_size, 1, infile))
        {
            int save_errno = errno;
            if (feof(infile))
            {
                fprintf(stderr, "fread of payload: unexpected EOF\n");
            }
            else
            {
                errno = save_errno;
                perror("fread of payload");
            }
            return 1;
        }

        /* check that padding bits are 0 */
        if (payload_bits &&
            (payload_buf[read_size-1] & ~((1<<payload_bits)-1)))
        {
            fprintf(stderr, "nonzero padding at end of bitstream\n");
            return 1;
        }

        /* dump */
#if 0
        if (block_count == 0)
        {
            FILE *outfile = fopen("dump.bin", "wb");

            fwrite(payload_buf, read_size, 1, outfile);
            fclose(outfile);
        }
#endif

        block_count ++;
    }

    if (block_count * block_mult != nominal_size)
    {
        fprintf(stderr, "size mismatch\n");
        return 1;
    }

    printf("ok!\n");

    return 0;
}
